// Iostreams base classes -*- C++ -*-

// Copyright (C) 1997-1999 Cygnus Solutions
//
// This file is part of the GNU ISO C++ Library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2, or (at your option)
// any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License along
// with this library; see the file COPYING.  If not, write to the Free
// Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.

// As a special exception, you may use this file as part of a free software
// library without restriction.  Specifically, if other files instantiate
// templates or use macros or inline functions from this file, or you compile
// this file and link it with other files to produce an executable, this
// file does not by itself cause the resulting executable to be covered by
// the GNU General Public License.  This exception does not however
// invalidate any other reasons why the executable file might be covered by
// the GNU General Public License.

//
// ISO C++ 14882: 27.4  Iostreams base classes
//

#include <bits/std_ios.h>
#include <bits/std_ostream.h>
#include <bits/std_istream.h>
#include <bits/std_fstream.h>
#include <bits/std_iostream.h>

namespace std {

  int ios_base::Init::_M_ios_base_init = 0;

  ios_base::Init::Init()
  {
    if (++_M_ios_base_init == 1)
      {
	_M_cin = new filebuf(0, "stdin", ios_base::in);
       	_M_cout = new filebuf(1, "stdout", ios_base::out);
	_M_cerr = new filebuf(2, "stderr", ios_base::out);
	cin.rdbuf(_M_cin);
	cout.rdbuf(_M_cout);
	cerr.rdbuf(_M_cerr);
	clog.rdbuf(_M_cerr);
#ifdef _G_USE_WCHAR_T
	_M_wcin = new wfilebuf(0, "stdin", ios_base::in);
	_M_wcout = new wfilebuf(1, "stdout", ios_base::out);
	_M_wcerr = new wfilebuf(2, "stderr", ios_base::out);
	wcin.rdbuf(_M_wcin);
	wcout.rdbuf(_M_wcout);
	wcerr.rdbuf(_M_wcerr);
	wclog.rdbuf(_M_wcerr);
#endif
      }
  }

  ios_base::Init::~Init()
  {
    if (--_M_ios_base_init == 0)
      {
	cout.flush();
	cerr.flush();
	clog.flush();
	delete _M_cin;
	delete _M_cout;
	delete _M_cerr;
	_M_cin = NULL;
	_M_cout = NULL;
	_M_cerr = NULL;
#ifdef _G_USE_WCHAR_T
	wcout.flush();
	wcerr.flush();
	wclog.flush();
	delete _M_wcin;
	delete _M_wcout;
	delete _M_wcerr;
	_M_wcin = NULL;
	_M_wcout = NULL;
	_M_wcerr = NULL;
#endif
      }
  } 

  // 27.4.2.5  ios_base storage functions
  int 
  ios_base::xalloc() throw()
  {
    // XXX MT
    // XXX should be a symbol. (Reserve 0..3 for builtins.)
    static int top = 4; 
    return top++;
  }

  // 27.4.2.5  iword/pword storage
  ios_base::_Words&
  ios_base::_M_grow_words(int ix)
  {
    // Precondition: _M_word_limit <= ix
    _Words zero = { 0, 0 };
    int newlimit = _S_local_words;
    _Words* words = _M_word_array;
    int i = 0;
    if (_S_local_words <= ix)
      {
	newlimit = ix+1;
	try
	  { words = new _Words[ix+1]; }
	catch (...)
	  {
	    _M_dummy = zero;  // XXX MT? Not on "normal" machines.
	    _M_clear(_M_rdstate() | badbit);  // may throw
	    return _M_dummy;
	  }
	do { words[i] = _M_words[i]; } while (++i < _M_word_limit);
	if (_M_words != _M_word_array) delete [] _M_words;
      }
    
    do { words[i] = zero; } while (++i < newlimit);
    _M_words = words;
    _M_word_limit = newlimit;
    return words[ix];
  }

  // 27.4.2.6  ios_base callbacks
  struct ios_base::_Callback_list
  {
    // Data Members
    _Callback_list* 		_M_next;
    ios_base::event_callback 	_M_fn;
    int 			_M_index;
    int 			_M_refcount;  // 0 means one reference.
    
    _Callback_list(ios_base::event_callback fn, 
		   int index, _Callback_list* __cb)
    : _M_next(__cb), _M_fn(fn), _M_index(index), _M_refcount(0) { }

    void 
    _M_add_reference() { ++_M_refcount; } // XXX MT

    int 
    _M_remove_reference() { return _M_refcount--; }  // 0 => OK to delete
  };

  void 
  ios_base::register_callback(event_callback fn, int index)
  { _M_callbacks = new _Callback_list(fn, index, _M_callbacks); }

  void 
  ios_base::_M_call_callbacks(event ev) throw()
  {
    for (_Callback_list* run = _M_callbacks; run; run = run->_M_next)
      {
	try { 
	  (*run->_M_fn) (ev, *this, run->_M_index); 
	} 
	catch (...) { 
	}
      }
  }

  void 
  ios_base::_M_dispose_callbacks()
  {
    _Callback_list* __p = _M_callbacks;
    while (__p && __p->_M_remove_reference() == 0)
      {
	_Callback_list* __next = __p->_M_next;
	delete __p;
	__p = __next;
      }
    _M_callbacks = 0;
  }
  
  // Called only by basic_ios<>::init.
  void 
  ios_base::_M_init()   
  {
    // XXX may be called more than once
    _M_flags = skipws | dec;
    _M_width = 0;
    _M_precision = 6;
    _M_callbacks = 0;
    _M_words = 0;
    _M_word_limit = 0;
    _M_state = goodbit;
    _M_exceptions_data = goodbit;
    _M_locale = locale();
    // no init needed for _M_word_array or _M_dummy.
  }
  
  void 
  ios_base::_M_copy_base(ios_base& __rhs)
  {
    // Alloc any new word array first, so if it fails we have "rollback".
    _Words* __words = (__rhs._M_word_limit <= _S_local_words) ?
      _M_word_array : new _Words[__rhs._M_word_limit];
    // Bump refs before doing callbacks, for safety.
    _Callback_list* __cb = __rhs._M_callbacks;
    if (__cb) 
      __cb->_M_add_reference();
    
    _M_call_callbacks (erase_event);
    
    if (_M_words != _M_word_array) 
      delete [] _M_words;
    _M_dispose_callbacks();
    _M_callbacks = __cb;  // NB: don't want any added during above.
    for (int __i = 0; __i < __rhs._M_word_limit; ++__i)
      __words[__i] = __rhs._M_words[__i];
    if (_M_words != _M_word_array) 
      delete [] _M_words;
    _M_words = __words;
    _M_word_limit = __rhs._M_word_limit;
#if 0
    // Per 27.1.1.1, does not call imbue(); but must trash all
    // caches associated with imbue(); must not imbue on the sbuf.
    _M_locale = __rhs._M_locale;
#endif
    flags (__rhs.flags());
    width (__rhs.width());
    precision (__rhs.precision());
  }
  
  // 27.4.2.3  ios_base locale functions
  locale
  ios_base::imbue(const locale& __loc)
  {
    locale __tmp = _M_locale;
    // Make sure there's a callback for the format caches so they will be
    // marked dirty.o
    _Format_cache<char>::_S_get(*this);
    _Format_cache<wchar_t>::_S_get(*this);
    _M_call_callbacks(imbue_event);
    _M_locale = __loc;
    // XXX not done?
    return __tmp;
  }

  ios_base::ios_base()
  {
    // Do nothing; init() does it.  Static init to 0 makes everything sane.
  }
  
  // 27.4.2.7  ios_base constructors/destructors
  ios_base::~ios_base()
  {
    _M_call_callbacks(erase_event);
    _M_dispose_callbacks();
    if (_M_words != _M_word_array) 
      delete [] _M_words;
    // XXX done?
  }

}  // namespace std









